from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple

import pyarrow as pa
from triad import Schema, assert_or_throw, to_uuid
from triad.utils.pyarrow import _type_to_expression
from triad.utils.schema import quote_name

from fugue.column.expressions import (
    ColumnExpr,
    _BinaryOpExpr,
    _FuncExpr,
    _LiteralColumnExpr,
    _NamedColumnExpr,
    _UnaryOpExpr,
    _WildcardExpr,
    col,
    lit,
)
from fugue.column.functions import is_agg
from fugue.exceptions import FugueBug

_SUPPORTED_OPERATORS: Dict[str, str] = {
    "+": "+",
    "-": "-",
    "*": "*",
    "/": "/",
    "&": " AND ",
    "|": " OR ",
    "<": "<",
    ">": ">",
    "<=": "<=",
    ">=": ">=",
    "==": "=",
    "!=": "!=",
}


class SelectColumns:
    """SQL ``SELECT`` columns collection.

    :param cols: collection of :class:`~fugue.column.expressions.ColumnExpr`
    :param arg_distinct: whether this is ``SELECT DISTINCT``, defaults to False

    .. admonition:: New Since
        :class: hint

        **0.6.0**
    """

    def __init__(self, *cols: ColumnExpr, arg_distinct: bool = False):  # noqa: C901
        self._distinct = arg_distinct
        self._all: List[ColumnExpr] = []
        self._literals: List[ColumnExpr] = []
        self._cols: List[ColumnExpr] = []
        self._non_agg_funcs: List[ColumnExpr] = []
        self._agg_funcs: List[ColumnExpr] = []
        self._group_keys: List[ColumnExpr] = []
        self._has_wildcard = False

        _g_keys: List[ColumnExpr] = []

        for c in cols:
            c = c.infer_alias()
            _is_agg = False
            self._all.append(c)
            if isinstance(c, _LiteralColumnExpr):
                self._literals.append(c)
            else:
                if isinstance(c, _NamedColumnExpr):
                    self._cols.append(c)
                elif isinstance(c, _WildcardExpr):
                    if self._has_wildcard:
                        raise ValueError("'*' can be used at most once")
                    self._has_wildcard = True
                    self._cols.append(c)
                elif isinstance(c, _FuncExpr):
                    if is_agg(c):
                        _is_agg = True
                        self._agg_funcs.append(c)
                    else:
                        self._non_agg_funcs.append(c)
                if not _is_agg:
                    if isinstance(c, _WildcardExpr):
                        _g_keys.append(c)
                    else:
                        _g_keys.append(c.alias("").cast(None))

        if self.has_agg:
            self._group_keys += _g_keys

        if len(self._agg_funcs) > 0 and self._has_wildcard:
            raise ValueError(f"'*' can't be used in aggregation: {self}")

    def __str__(self):
        """String representation for debug purpose"""
        expr = ", ".join(str(x) for x in self.all_cols)
        return f"[{expr}]"

    def __uuid__(self):
        """Unique id for this collection"""
        return to_uuid(self._distinct, self.all_cols)

    @property
    def is_distinct(self) -> bool:
        """Whether this is a ``SELECT DISTINCT``"""
        return self._distinct

    def replace_wildcard(self, schema: Schema) -> "SelectColumns":
        """Replace wildcard ``*`` with explicit column names

        :param schema: the schema used to parse the wildcard
        :return: a new instance containing only explicit columns

        .. note::

            It only replaces the top level ``*``. For example
            ``count_distinct(all_cols())`` will not be transformed because
            this ``*`` is not first level.
        """

        def _get_cols() -> Iterable[ColumnExpr]:
            for c in self.all_cols:
                if isinstance(c, _WildcardExpr):
                    yield from [col(n) for n in schema.names]
                else:
                    yield c

        return SelectColumns(*list(_get_cols()))

    def assert_all_with_names(self) -> "SelectColumns":
        """Assert every column have explicit alias or the alias can
        be inferred (non empty value). It will also validate there is
        no duplicated aliases

        :raises ValueError: if there are columns without alias, or there are
          duplicated names.
        :return: the instance itself
        """

        names: Set[str] = set()
        for x in self.all_cols:
            if isinstance(x, _WildcardExpr):
                continue
            if isinstance(x, _NamedColumnExpr):
                if self._has_wildcard:
                    if x.as_name == "":
                        raise ValueError(
                            f"with '*', all other columns must have an alias: {self}"
                        )
            if x.output_name == "":
                raise ValueError(f"{x} does not have an alias: {self}")
            if x.output_name in names:
                raise ValueError(f"{x} can't be reused in select: {self}")
            names.add(x.output_name)

        return self

    def assert_no_wildcard(self) -> "SelectColumns":
        """Assert there is no ``*`` on first level columns

        :raises AssertionError: if ``all_cols()`` exists
        :return: the instance itself
        """
        assert not self._has_wildcard
        return self

    def assert_no_agg(self) -> "SelectColumns":
        """Assert there is no aggregation operation on any column.

        :raises AssertionError: if there is any aggregation in the
          collection.
        :return: the instance itself

        .. seealso::
            Go to :func:`~fugue.column.functions.is_agg` to see how the
            aggregations are detected.
        """
        assert not self.has_agg
        return self

    @property
    def all_cols(self) -> List[ColumnExpr]:
        """All columns (with inferred aliases)"""
        return self._all

    @property
    def literals(self) -> List[ColumnExpr]:
        """All literal columns"""
        return self._literals

    @property
    def simple_cols(self) -> List[ColumnExpr]:
        """All columns directly representing column names"""
        return self._cols

    @property
    def non_agg_funcs(self) -> List[ColumnExpr]:
        """All columns with non-aggregation operations"""
        return self._non_agg_funcs

    @property
    def agg_funcs(self) -> List[ColumnExpr]:
        """All columns with aggregation operations"""
        return self._agg_funcs

    @property
    def group_keys(self) -> List[ColumnExpr]:
        """Group keys inferred from the columns.

        .. note::

            * if there is no aggregation, the result will be empty
            * it is :meth:`~.simple_cols` plus :meth:`~.non_agg_funcs`
        """
        return self._group_keys

    @property
    def has_agg(self) -> bool:
        """Whether this select is an aggregation"""
        return len(self.agg_funcs) > 0

    @property
    def has_literals(self) -> bool:
        """Whether this select contains literal columns"""
        return len(self.literals) > 0

    @property
    def simple(self) -> bool:
        """Whether this select contains only simple column representations"""
        return len(self.simple_cols) == len(self.all_cols)


class SQLExpressionGenerator:
    """SQL generator for :class:`~.SelectColumns`

    :param enable_cast: whether convert ``cast`` into the statement, defaults to True

    .. admonition:: New Since
        :class: hint

        **0.6.0**
    """

    def __init__(self, enable_cast: bool = True):
        self._enable_cast = enable_cast
        self._func_handler: Dict[str, Callable[[_FuncExpr], Iterable[str]]] = {}

    def where(self, condition: ColumnExpr, table: str) -> Iterable[Tuple[bool, str]]:
        """Generate a ``SELECT *`` statement with the given where clause

        :param condition: column expression for ``WHERE``
        :param table: table name for ``FROM``
        :return: the SQL statement

        :raises ValueError: if ``condition`` contains aggregation

        .. admonition:: Examples

            .. code-block:: python

                gen = SQLExpressionGenerator(enable_cast=False)

                # SELECT * FROM tb WHERE a>1 AND b IS NULL
                gen.where((col("a")>1) & col("b").is_null(), "tb")
        """
        assert_or_throw(
            not is_agg(condition),
            lambda: ValueError(f"{condition} has aggregation functions"),
        )
        cond = self.generate(condition.alias(""))
        yield (False, "SELECT * FROM")
        yield (True, table)
        yield (False, f"WHERE {cond}")

    def select(
        self,
        columns: SelectColumns,
        table: str,
        where: Optional[ColumnExpr] = None,
        having: Optional[ColumnExpr] = None,
    ) -> Iterable[Tuple[bool, str]]:
        """Construct the full ``SELECT`` statement on a single table

        :param columns: columns to select, it may contain aggregations, if
          so, the group keys are inferred.
          See :meth:`~fugue.column.sql.SelectColumns.group_keys`
        :param table: table name to select from
        :param where: ``WHERE`` condition, defaults to None
        :param having: ``HAVING`` condition, defaults to None. It is used
          only when there is aggregation
        :return: the full ``SELECT`` statement
        """
        columns.assert_all_with_names()

        def _where() -> str:
            if where is None:
                return ""
            assert_or_throw(
                not is_agg(where),
                lambda: ValueError(f"{where} has aggregation functions"),
            )
            return "WHERE " + self.generate(where.alias(""))

        def _having(as_where: bool = False) -> str:
            if having is None:
                return ""
            pre = "WHERE " if as_where else "HAVING "
            return pre + self.generate(having.alias(""))

        distinct = "" if not columns.is_distinct else "DISTINCT "

        if not columns.has_agg:
            expr = ", ".join(self.generate(x) for x in columns.all_cols)
            yield (False, f"SELECT {distinct}{expr} FROM")
            yield (True, table)
            yield (False, _where())
            return
        columns.assert_no_wildcard()
        if len(columns.literals) == 0:
            expr = ", ".join(self.generate(x) for x in columns.all_cols)
            if len(columns.group_keys) == 0:
                yield (False, f"SELECT {distinct}{expr} FROM")
                yield (True, table)
                yield (False, _where())
                yield (False, _having())
                return
            else:
                keys = ", ".join(self.generate(x) for x in columns.group_keys)
                yield (False, f"SELECT {distinct}{expr} FROM")
                yield (True, table)
                yield (False, _where())
                yield (False, f"GROUP BY {keys}")
                yield (False, _having())
                return
        else:
            no_lit = [
                x for x in columns.all_cols if not isinstance(x, _LiteralColumnExpr)
            ]
            sub = self.select(SelectColumns(*no_lit), table, where=where, having=having)
            names = [
                self.generate(x) if isinstance(x, _LiteralColumnExpr) else x.output_name
                for x in columns.all_cols
            ]
            expr = ", ".join(names)
            yield (False, f"SELECT {expr} FROM (")
            yield from sub
            yield (False, ")")

    def generate(self, expr: ColumnExpr) -> str:
        """Convert :class:`~fugue.column.expressions.ColumnExpr` to
        SQL clause

        :param expr: the column expression to convert
        :return: the SQL clause for this expression
        """
        return "".join(self._generate(expr)).strip()

    def add_func_handler(
        self, name: str, handler: Callable[[_FuncExpr], Iterable[str]]
    ) -> "SQLExpressionGenerator":
        """Add special function handler.

        :param name: name of the function
        :param handler: the function to convert the function expression to SQL
          clause
        :return: the instance itself

        .. caution::

            Users should not use this directly
        """
        self._func_handler[name] = handler
        return self

    def correct_select_schema(
        self, input_schema: Schema, select: SelectColumns, output_schema: Schema
    ) -> Optional[Schema]:
        """Do partial schema inference from ``input_schema`` and ``select`` columns,
        then compare with the SQL output dataframe schema, and return the different
        part as a new schema, or None if there is no difference

        :param input_schema: input dataframe schema for the select statement
        :param select: the collection of select columns
        :param output_schema: schema of the output dataframe after executing the SQL
        :return: the difference as a new schema or None if no difference

        .. tip::

            This is particularly useful when the SQL engine messed up the schema of the
            output. For example, ``SELECT *`` should return a dataframe with the same
            schema of the input. However, for example a column ``a:int`` could become
            ``a:long`` in the output dataframe because of information loss. This
            function is designed to make corrections on column types when they can be
            inferred. This may not be perfect but it can solve major discrepancies.
        """
        cols = select.replace_wildcard(input_schema).assert_all_with_names()
        fields: List[pa.Field] = []
        for c in cols.all_cols:
            tp = c.infer_type(input_schema)
            if tp is not None and tp != output_schema[c.output_name].type:
                fields.append(pa.field(c.output_name, tp))
        if len(fields) == 0:
            return None
        return Schema(fields)

    def _generate(  # noqa: C901
        self, expr: ColumnExpr, bracket: bool = False
    ) -> Iterable[str]:
        if self._enable_cast and expr.as_type is not None:
            yield "CAST("
        if isinstance(expr, _LiteralColumnExpr):
            yield from self._on_lit(expr)
        elif isinstance(expr, _NamedColumnExpr):
            yield from self._on_named(expr)
        elif isinstance(expr, _WildcardExpr):
            yield "*"
        elif isinstance(expr, _FuncExpr):
            if expr.func in self._func_handler:
                yield from self._func_handler[expr.func](expr)
            elif isinstance(expr, _UnaryOpExpr):
                yield from self._on_common_unary(expr)
            elif isinstance(expr, _BinaryOpExpr):
                yield from self._on_common_binary(expr, bracket)
            else:
                yield from self._on_common_func(expr)
        if self._enable_cast and expr.as_type is not None:
            yield " AS "
            yield self.type_to_expr(expr.as_type)
            yield ")"
        if expr.as_name != "":
            yield " AS " + quote_name(expr.as_name)
        elif expr.as_type is not None and expr.name != "":
            yield " AS " + quote_name(expr.name)

    def type_to_expr(self, data_type: pa.DataType):
        return _type_to_expression(data_type)

    def _on_named(self, expr: _NamedColumnExpr) -> Iterable[str]:
        yield quote_name(expr.name)

    def _on_lit(self, expr: _LiteralColumnExpr) -> Iterable[str]:
        yield expr.body_str

    def _on_common_unary(self, expr: _UnaryOpExpr) -> Iterable[str]:
        if expr.op == "-":
            yield expr.op
            yield from self._generate(expr.col, bracket=True)
        elif expr.op == "~":
            yield "NOT "
            yield from self._generate(expr.col, bracket=True)
        elif expr.op == "IS_NULL":
            yield from self._generate(expr.col, bracket=True)
            yield " IS NULL"
        elif expr.op == "NOT_NULL":
            yield from self._generate(expr.col, bracket=True)
            yield " IS NOT NULL"
        else:
            raise NotImplementedError(expr)  # pragma: no cover

    def _on_common_binary(self, expr: _BinaryOpExpr, bracket: bool) -> Iterable[str]:
        assert_or_throw(expr.op in _SUPPORTED_OPERATORS, NotImplementedError(expr))
        if bracket:
            yield "("
        if expr.is_distinct:  # pragma: no cover
            raise FugueBug(f"impossible case {expr}")
        yield from self._generate(expr.left, bracket=True)
        yield _SUPPORTED_OPERATORS[expr.op]
        yield from self._generate(expr.right, bracket=True)
        if bracket:
            yield ")"

    def _on_common_func(self, expr: _FuncExpr) -> Iterable[str]:
        def to_str(v: Any) -> Iterable[str]:
            if isinstance(v, ColumnExpr):
                yield from self._generate(v)
            else:
                yield from self._generate(lit(v))

        def get_args() -> Iterable[str]:
            for x in expr.args:
                yield from to_str(x)
                yield ","
            for k, v in expr.kwargs.items():
                yield k
                yield "="
                yield from to_str(v)
                yield ","

        args = list(get_args())
        if len(args) > 0:
            args = args[:-1]
        yield expr.func
        yield "("
        if expr.is_distinct:
            yield "DISTINCT "
        yield from args
        yield ")"



from contextlib import contextmanager
from typing import Any, Dict, Iterator

import ray

import fugue.test as ft


@ft.fugue_test_backend
class RayTestBackend(ft.FugueTestBackend):
    name = "ray"
    default_session_conf = {"num_cpus": 2}
    default_fugue_conf = {
        "fugue.ray.zero_copy": True,
        "fugue.ray.default.batch_size": 10000,
    }

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        with ray.init(**session_conf):
            yield "ray"



import math
from typing import Any, Callable, List, Optional, Tuple, TypeVar

import dask.dataframe as dd
import numpy as np
import pandas as pd
import pyarrow as pa
from dask.dataframe import DataFrame
from dask.delayed import delayed
from dask.distributed import Client, get_client
from triad.utils.pandas_like import PD_UTILS, PandasLikeUtils
from triad.utils.pyarrow import to_pandas_dtype

import fugue.api as fa
from fugue.constants import FUGUE_CONF_DEFAULT_PARTITIONS

from ._constants import FUGUE_DASK_CONF_DEFAULT_PARTITIONS

_FUGUE_DASK_TEMP_IDX_COLUMN = "_fugue_dask_temp_index"
T = TypeVar("T")


def get_default_partitions() -> int:
    """Get the default number of partitions"""
    conf = fa.get_current_conf()
    n = conf.get(
        FUGUE_DASK_CONF_DEFAULT_PARTITIONS,
        conf.get(FUGUE_CONF_DEFAULT_PARTITIONS, -1),
    )
    return n if n > 0 else fa.get_current_parallelism() * 2


def collect(df: dd.DataFrame, func: Callable[[pd.DataFrame], T]) -> Tuple[T]:
    """Compute each partition in parallel and collect the results

    :param df: dask dataframe
    :return: the collected result
    """
    dfs = df.to_delayed()
    objs = [delayed(func)(df) for df in dfs]
    return dd.compute(*objs)


def hash_repartition(df: dd.DataFrame, num: int, cols: List[Any]) -> dd.DataFrame:
    """Repartition the dataframe by hashing the given columns

    :param df: dataframe to repartition
    :param num: number of partitions
    :param cols: columns to hash, if empty, all columns will be used

    :return: repartitioned dataframe
    """
    if num < 1:
        return df
    if num == 1:
        return df.repartition(npartitions=1)
    df = df.reset_index(drop=True).clear_divisions()
    idf, ct = _add_hash_index(df, num, cols)
    return _postprocess(idf, ct, num)


def even_repartition(df: dd.DataFrame, num: int, cols: List[Any]) -> dd.DataFrame:
    """Repartition the dataframe by evenly distributing the given columns

    :param df: dataframe to repartition
    :param num: number of partitions
    :param cols: group columns

    :return: repartitioned dataframe

    .. note::

        When cols is empty, the dataframe will be evenly repartitioned by ``num``.
        When cols is not empty, the dataframe will be evenly repartitioned by the
        groups of the given columns. When cols is not empty, if ``num<=0``,
        the number of partitions will be the number of groups.
    """
    if num == 1:
        return df.repartition(npartitions=1)
    if len(cols) == 0 and num <= 0:
        return df
    df = df.reset_index(drop=True).clear_divisions()
    if len(cols) == 0:
        idf, ct = _add_continuous_index(df)
    else:
        idf, ct = _add_group_index(df, cols, shuffle=False)
        # when cols are set and num is not set, we use the number of groups
        if num <= 0:
            num = ct
    return _postprocess(idf, ct, num)


def rand_repartition(
    df: dd.DataFrame, num: int, cols: List[Any], seed: Any = None
) -> dd.DataFrame:
    """Repartition the dataframe by randomly distributing the rows or groups

    :param df: dataframe to repartition
    :param num: number of partitions
    :param cols: group columns
    :param seed: random seed

    :return: repartitioned dataframe

    .. note::

        When cols is empty, the dataframe will be randomly shuffled and
        repartitioned by ``num``. When cols is not empty, the dataframe will be
        grouped and then randomly shuffled by groups.
    """
    if num < 1:
        return df
    if num == 1:
        return df.repartition(npartitions=1)
    df = df.reset_index(drop=True).clear_divisions()
    if len(cols) == 0:
        idf, ct = _add_random_index(df, num=num, seed=seed)
    else:
        idf, ct = _add_group_index(df, cols, shuffle=True, seed=seed)
        # when cols are set and num is not set, we use the number of groups
    return _postprocess(idf, ct, num)


def _postprocess(idf: dd.DataFrame, ct: int, num: int) -> dd.DataFrame:
    parts = min(ct, num)
    if parts <= 1:
        return idf.repartition(npartitions=1)
    divisions = list(np.arange(ct, step=math.ceil(ct / parts)))
    divisions.append(ct - 1)
    return idf.repartition(divisions=divisions, force=True)


def _add_group_index(
    df: dd.DataFrame, cols: List[str], shuffle: bool, seed: Any = None
) -> Tuple[dd.DataFrame, int]:
    keys = df[cols].drop_duplicates().compute()
    if shuffle:
        keys = keys.sample(frac=1, random_state=seed)
    keys = keys.reset_index(drop=True).assign(
        **{_FUGUE_DASK_TEMP_IDX_COLUMN: pd.Series(range(len(keys)), dtype=int)}
    )
    df = df.merge(dd.from_pandas(keys, npartitions=1), on=cols, broadcast=True)
    return df.set_index(_FUGUE_DASK_TEMP_IDX_COLUMN, drop=True), len(keys)


def _add_hash_index(
    df: dd.DataFrame, num: int, cols: List[str]
) -> Tuple[dd.DataFrame, int]:
    if len(cols) == 0:
        cols = list(df.columns)

    def _add_hash(df: pd.DataFrame) -> pd.DataFrame:  # pragma: no cover
        if len(df) == 0:
            return df.assign(**{_FUGUE_DASK_TEMP_IDX_COLUMN: pd.Series(dtype=int)})
        return df.assign(
            **{
                _FUGUE_DASK_TEMP_IDX_COLUMN: pd.util.hash_pandas_object(
                    df[cols], index=False
                )
                .mod(num)
                .astype(int)
            }
        )

    orig_schema = list(df.dtypes.to_dict().items())
    idf = df.map_partitions(
        _add_hash, meta=orig_schema + [(_FUGUE_DASK_TEMP_IDX_COLUMN, int)]
    ).set_index(_FUGUE_DASK_TEMP_IDX_COLUMN, drop=True)
    return idf, num


def _add_random_index(
    df: dd.DataFrame, num: int, seed: Any = None
) -> Tuple[dd.DataFrame, int]:  # pragma: no cover
    def _add_rand(df: pd.DataFrame) -> pd.DataFrame:
        if len(df) == 0:
            return df.assign(**{_FUGUE_DASK_TEMP_IDX_COLUMN: pd.Series(dtype=int)})
        if seed is not None:
            np.random.seed(seed)
        return df.assign(
            **{_FUGUE_DASK_TEMP_IDX_COLUMN: np.random.randint(0, num, len(df))}
        )

    orig_schema = list(df.dtypes.to_dict().items())
    idf = df.map_partitions(
        _add_rand, meta=orig_schema + [(_FUGUE_DASK_TEMP_IDX_COLUMN, int)]
    ).set_index(_FUGUE_DASK_TEMP_IDX_COLUMN, drop=True)
    return idf, num


def _add_continuous_index(df: dd.DataFrame) -> Tuple[dd.DataFrame, int]:
    def _get_info(
        df: pd.DataFrame, partition_info: Any
    ) -> pd.DataFrame:  # pragma: no cover
        return pd.DataFrame(dict(no=[partition_info["number"]], ct=[len(df)]))

    pinfo = (
        df.index.to_frame(name=df.index.name)
        .map_partitions(_get_info, meta={"no": int, "ct": int})
        .compute()
    )
    counts = pinfo.sort_values("no").ct.cumsum().tolist()
    starts = [0] + counts[0:-1]

    def _add_index(
        df: pd.DataFrame, partition_info: Any
    ) -> pd.DataFrame:  # pragma: no cover
        return df.assign(
            **{
                _FUGUE_DASK_TEMP_IDX_COLUMN: np.arange(len(df))
                + starts[partition_info["number"]]
            }
        )

    orig_schema = list(df.dtypes.to_dict().items())
    idf = df.map_partitions(
        _add_index, meta=orig_schema + [(_FUGUE_DASK_TEMP_IDX_COLUMN, int)]
    )
    idf = idf.set_index(_FUGUE_DASK_TEMP_IDX_COLUMN, drop=True)
    return idf, counts[-1]


class DaskUtils(PandasLikeUtils[dd.DataFrame, dd.Series]):
    def concat_dfs(self, *dfs: dd.DataFrame) -> dd.DataFrame:
        return dd.concat(list(dfs))

    def get_or_create_client(self, client: Optional[Client] = None):
        if client is not None:
            return client
        try:
            return get_client()
        except ValueError:  # pragma: no cover
            return Client(processes=True)

    def ensure_compatible(self, df: DataFrame) -> None:
        if df.index.name != _FUGUE_DASK_TEMP_IDX_COLUMN:
            super().ensure_compatible(df)

    def is_compatile_index(self, df: dd.DataFrame) -> bool:
        """Check whether the datafame is compatible with the operations inside
        this utils collection

        :param df: dask dataframe
        :return: if it is compatible
        """
        if isinstance(df.index, dd.Index):
            return True
        return isinstance(df.index, pd.RangeIndex) or (  # pragma: no cover
            hasattr(df.index, "inferred_type") and df.index.inferred_type == "integer"
        )

    def cast_df(
        self,
        df: DataFrame,
        schema: pa.Schema,
        use_extension_types: bool = True,
        use_arrow_dtype: bool = False,
        **kwargs: Any
    ) -> DataFrame:
        output_dtypes = to_pandas_dtype(
            schema,
            use_extension_types=use_extension_types,
            use_arrow_dtype=use_arrow_dtype,
        )
        return df.map_partitions(
            PD_UTILS.cast_df,
            schema=schema,
            use_extension_types=use_extension_types,
            use_arrow_dtype=use_arrow_dtype,
            meta=output_dtypes,
            **kwargs
        )


DASK_UTILS = DaskUtils()



from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
from triad.collections.schema import Schema
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_or_throw
from triad.utils.pyarrow import cast_pa_table, pa_table_to_pandas

from fugue.dataset.api import (
    as_fugue_dataset,
    as_local,
    as_local_bounded,
    count,
    get_num_partitions,
    is_bounded,
    is_empty,
    is_local,
)
from fugue.exceptions import FugueDataFrameOperationError

from .api import (
    alter_columns,
    as_array,
    as_array_iterable,
    as_dict_iterable,
    as_dicts,
    as_pandas,
    drop_columns,
    get_column_names,
    get_schema,
    is_df,
    rename,
    select_columns,
)
from .dataframe import DataFrame, LocalBoundedDataFrame, _input_schema
from .utils import (
    pa_table_as_array,
    pa_table_as_array_iterable,
    pa_table_as_dict_iterable,
    pa_table_as_dicts,
)


class ArrowDataFrame(LocalBoundedDataFrame):
    """DataFrame that wraps :func:`pyarrow.Table <pa:pyarrow.table>`. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: 2-dimensional array, iterable of arrays,
      :func:`pyarrow.Table <pa:pyarrow.table>` or pandas DataFrame
    :param schema: |SchemaLikeObject|

    .. admonition:: Examples

        >>> ArrowDataFrame([[0,'a'],[1,'b']],"a:int,b:str")
        >>> ArrowDataFrame(schema = "a:int,b:int")  # empty dataframe
        >>> ArrowDataFrame(pd.DataFrame([[0]],columns=["a"]))
        >>> ArrowDataFrame(ArrayDataFrame([[0]],"a:int).as_arrow())
    """

    def __init__(  # noqa: C901
        self,
        df: Any = None,
        schema: Any = None,
    ):
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            self._native: pa.Table = schema.create_empty_arrow_table()
            super().__init__(schema)
            return
        elif isinstance(df, pa.Table):
            assert_or_throw(
                schema is None,
                InvalidOperationError("can't reset schema for pa.Table"),
            )
            self._native = df
            super().__init__(Schema(df.schema))
            return
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if isinstance(df, pd.Series):
                df = df.to_frame()
            pdf = df
            if schema is None:
                self._native = pa.Table.from_pandas(
                    pdf,
                    schema=Schema(pdf).pa_schema,
                    preserve_index=False,
                    safe=True,
                )
                schema = Schema(self._native.schema)
            else:
                schema = _input_schema(schema).assert_not_empty()
                if pdf.shape[0] == 0:
                    self._native = schema.create_empty_arrow_table()
                else:
                    self._native = pa.Table.from_pandas(
                        pdf,
                        schema=schema.pa_schema,
                        preserve_index=False,
                        safe=True,
                    )
            super().__init__(schema)
            return
        elif isinstance(df, Iterable):
            schema = _input_schema(schema).assert_not_empty()
            pdf = pd.DataFrame(df, columns=schema.names)
            if pdf.shape[0] == 0:
                self._native = schema.create_empty_arrow_table()
            else:
                for f in schema.fields:
                    if pa.types.is_timestamp(f.type) or pa.types.is_date(f.type):
                        pdf[f.name] = pd.to_datetime(pdf[f.name])
                schema = _input_schema(schema).assert_not_empty()
                self._native = pa.Table.from_pandas(
                    pdf, schema=schema.pa_schema, preserve_index=False, safe=True
                )
            super().__init__(schema)
            return
        else:
            raise ValueError(f"{df} is incompatible with ArrowDataFrame")

    @property
    def native(self) -> pa.Table:
        """:func:`pyarrow.Table <pa:pyarrow.table>`"""
        return self._native

    def native_as_df(self) -> pa.Table:
        return self._native

    @property
    def empty(self) -> bool:
        return self.count() == 0

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        data = self.native.take([0]).to_pydict()
        return [v[0] for v in data.values()]

    def peek_dict(self) -> Dict[str, Any]:
        self.assert_not_empty()
        data = self.native.take([0]).to_pydict()
        return {k: v[0] for k, v in data.items()}

    def count(self) -> int:
        return self.native.shape[0]

    def as_pandas(self) -> pd.DataFrame:
        return _pa_table_as_pandas(self.native)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        adf = self.native if columns is None else self.native.select(columns)
        n = min(n, self.count())
        if n == 0:
            schema = self.schema if columns is None else self.schema.extract(columns)
            return ArrowDataFrame(None, schema=schema)
        return ArrowDataFrame(adf.take(list(range(n))))

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        return ArrowDataFrame(self.native.drop(cols))

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        return ArrowDataFrame(self.native.select(keys))

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            cols = dict(columns)
            new_cols = [cols.pop(c, c) for c in self.columns]
            assert_or_throw(len(cols) == 0)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        return ArrowDataFrame(self.native.rename_columns(new_cols))

    def alter_columns(self, columns: Any) -> DataFrame:
        adf = _pa_table_alter_columns(self.native, columns)
        if adf is self.native:
            return self
        return ArrowDataFrame(adf)

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return self.native

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return pa_table_as_array(self.native, columns=columns)

    def as_dicts(self, columns: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return pa_table_as_dicts(self.native, columns=columns)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        yield from pa_table_as_array_iterable(self.native, columns=columns)

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        yield from pa_table_as_dict_iterable(self.native, columns=columns)


@as_local.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_as_local(df: pa.Table) -> pa.Table:
    return df


@as_local_bounded.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_as_local_bounded(df: pa.Table) -> pa.Table:
    return df


@as_pandas.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_as_pandas(df: pa.Table) -> pd.DataFrame:
    return pa_table_to_pandas(
        df,
        use_extension_types=True,
        use_arrow_dtype=False,
        use_threads=False,
        date_as_object=False,
    )


@as_array.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _pa_table_as_array(
    df: pa.Table, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Any]:
    return pa_table_as_array(df, columns=columns)


@as_array_iterable.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _pa_table_as_array_iterable(
    df: pa.Table, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Any]:
    yield from pa_table_as_array_iterable(df, columns=columns)


@as_dicts.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _pa_table_as_dicts(
    df: pa.Table, columns: Optional[List[str]] = None
) -> List[Dict[str, Any]]:
    return pa_table_as_dicts(df, columns=columns)


@as_dict_iterable.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _pa_table_as_dict_iterable(
    df: pa.Table, columns: Optional[List[str]] = None
) -> Iterable[Dict[str, Any]]:
    yield from pa_table_as_dict_iterable(df, columns=columns)


@alter_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _pa_table_alter_columns(
    df: pa.Table, columns: Any, as_fugue: bool = False
) -> pa.Table:
    schema = Schema(df.schema)
    new_schema = schema.alter(columns)
    if schema != new_schema:
        df = cast_pa_table(df, new_schema.pa_schema)
    return df if not as_fugue else ArrowDataFrame(df)


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, pa.Table))
def _pa_table_as_fugue_df(df: pa.Table, **kwargs: Any) -> "ArrowDataFrame":
    return ArrowDataFrame(df, **kwargs)


@is_df.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_is_df(df: pa.Table) -> bool:
    return True


@count.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_count(df: pa.Table) -> int:
    return df.shape[0]


@is_bounded.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_is_bounded(df: pa.Table) -> bool:
    return True


@is_empty.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_is_empty(df: pa.Table) -> bool:
    return df.shape[0] == 0


@is_local.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_is_local(df: pa.Table) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, pa.Table))
def _pa_table_get_num_partitions(df: pa.Table) -> int:
    return 1


@get_column_names.candidate(lambda df: isinstance(df, pa.Table))
def _get_pyarrow_table_columns(df: pa.Table) -> List[Any]:
    return [f.name for f in df.schema]


@get_schema.candidate(lambda df: isinstance(df, pa.Table))
def _get_pyarrow_table_schema(df: pa.Table) -> Schema:
    return Schema(df.schema)


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _rename_pyarrow_dataframe(df: pa.Table, columns: Dict[str, Any]) -> pa.Table:
    if len(columns) == 0:
        return df
    _assert_no_missing(df, columns.keys())
    return df.rename_columns([columns.get(f.name, f.name) for f in df.schema])


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _drop_pa_columns(df: pa.Table, columns: List[str]) -> pa.Table:
    cols = [x for x in df.schema.names if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return df.select(cols)


@select_columns.candidate(lambda df, *args, **kwargs: isinstance(df, pa.Table))
def _select_pa_columns(df: pa.Table, columns: List[Any]) -> pa.Table:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns=columns)
    return df.select(columns)


def _build_empty_arrow(schema: Schema) -> pa.Table:  # pragma: no cover
    # TODO: remove
    return schema.create_empty_arrow_table()


def _assert_no_missing(df: pa.Table, columns: Iterable[Any]) -> None:
    missing = [x for x in columns if x not in df.schema.names]
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")



from typing import Any, Tuple

from duckdb import DuckDBPyConnection, DuckDBPyRelation
from triad import run_at_def

from fugue import (
    DataFrame,
    ExecutionEngine,
    register_execution_engine,
    register_sql_engine,
)
from fugue.dev import (
    DataFrameParam,
    ExecutionEngineParam,
    fugue_annotated_param,
    is_pandas_or,
)
from fugue.plugins import infer_execution_engine, parse_execution_engine
from fugue_duckdb.dataframe import DuckDataFrame
from fugue_duckdb.execution_engine import DuckDBEngine, DuckExecutionEngine

from .tester import DuckDBTestBackend  # noqa: F401  # pylint: disable-all

try:
    from .tester import DuckDaskTestBackend  # noqa: F401  # pylint: disable-all
except ImportError:  # pragma: no cover
    pass


@infer_execution_engine.candidate(
    lambda objs: is_pandas_or(objs, (DuckDBPyRelation, DuckDataFrame))
)
def _infer_duckdb_client(objs: Any) -> Any:
    return "duckdb"


def _register_engines() -> None:
    register_execution_engine(
        "duck",
        lambda conf, **kwargs: DuckExecutionEngine(conf=conf),
        on_dup="ignore",
    )
    register_execution_engine(
        "duckdb",
        lambda conf, **kwargs: DuckExecutionEngine(conf=conf),
        on_dup="ignore",
    )
    try:
        from fugue_duckdb.dask import DuckDaskExecutionEngine

        register_execution_engine(
            "duckdask",
            lambda conf, **kwargs: DuckDaskExecutionEngine(conf=conf),
            on_dup="ignore",
        )
        register_execution_engine(
            "duckdbdask",
            lambda conf, **kwargs: DuckDaskExecutionEngine(conf=conf),
            on_dup="ignore",
        )
        register_execution_engine(
            "dd",
            lambda conf, **kwargs: DuckDaskExecutionEngine(conf=conf),
            on_dup="ignore",
        )
    except Exception:  # pragma: no cover
        pass
    register_execution_engine(
        DuckDBPyConnection,
        lambda con, conf, **kwargs: DuckExecutionEngine(conf=conf, connection=con),
        on_dup="ignore",
    )
    register_sql_engine("duck", lambda engine: DuckDBEngine(engine))
    register_sql_engine("duckdb", lambda engine: DuckDBEngine(engine))


try:
    from fugue_duckdb.dask import DuckDaskExecutionEngine
    from dask.distributed import Client

    @parse_execution_engine.candidate(
        lambda engine, conf, **kwargs: isinstance(engine, list)
        and len(engine) == 2
        and isinstance(engine[0], DuckDBPyConnection)
        and isinstance(engine[1], Client),
    )
    def _parse_duck_dask_client(
        engine: Tuple[DuckDBPyConnection, Client], conf: Any, **kwargs: Any
    ) -> DuckDaskExecutionEngine:
        return DuckDaskExecutionEngine(
            connection=engine[0], dask_client=engine[1], conf=conf
        )

except Exception:  # pragma: no cover
    pass


@fugue_annotated_param(DuckExecutionEngine)
class _DuckExecutionEngineParam(ExecutionEngineParam):
    pass


@fugue_annotated_param(DuckDBPyConnection)
class _DuckDBPyConnectionParam(ExecutionEngineParam):
    def to_input(self, engine: ExecutionEngine) -> Any:
        assert isinstance(engine, DuckExecutionEngine)
        return engine.connection  # type:ignore


@fugue_annotated_param(DuckDBPyRelation)
class _DuckDBPyRelationParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        assert isinstance(ctx, DuckExecutionEngine)
        return ctx.to_df(df).native  # type: ignore

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, DuckDBPyRelation)
        assert isinstance(ctx, DuckExecutionEngine)
        return DuckDataFrame(output)

    def count(self, df: Any) -> int:  # pragma: no cover
        return df.count()


@run_at_def
def _register() -> None:
    """Register DuckDB Execution Engine

    .. note::

        This function is automatically called when you do

        >>> import fugue_duckdb
    """
    _register_engines()



from typing import Dict, Any

FUGUE_CONTRIB: Dict[str, Any] = {
    "viz": {"module": "fugue_contrib.viz"},
    "sns": {"module": "fugue_contrib.seaborn"},
    "why": {"module": "whylogs.api.fugue.registry"},
    "vizzu": {"module": "ipyvizzu.integrations.fugue"},
}



from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
from duckdb import DuckDBPyRelation
from triad import Schema, assert_or_throw
from triad.utils.pyarrow import LARGE_TYPES_REPLACEMENT, replace_types_in_table

from fugue import ArrowDataFrame, DataFrame, LocalBoundedDataFrame
from fugue.dataframe.arrow_dataframe import _pa_table_as_pandas
from fugue.dataframe.utils import (
    pa_table_as_array,
    pa_table_as_array_iterable,
    pa_table_as_dict_iterable,
    pa_table_as_dicts,
)
from fugue.exceptions import FugueDataFrameOperationError, FugueDatasetEmptyError
from fugue.plugins import (
    as_array,
    as_array_iterable,
    as_arrow,
    as_dict_iterable,
    as_dicts,
    as_fugue_dataset,
    as_local_bounded,
    as_pandas,
    drop_columns,
    get_column_names,
    get_num_partitions,
    get_schema,
    is_df,
    select_columns,
)

from ._utils import encode_column_name, to_duck_type, to_pa_type


class DuckDataFrame(LocalBoundedDataFrame):
    """DataFrame that wraps DuckDB ``DuckDBPyRelation``.

    :param rel: ``DuckDBPyRelation`` object
    """

    def __init__(self, rel: DuckDBPyRelation):
        self._rel = rel
        super().__init__(schema=lambda: _duck_get_schema(self._rel))

    @property
    def alias(self) -> str:
        return "_" + str(id(self._rel))  # DuckDBPyRelation.alias is not always unique

    @property
    def native(self) -> DuckDBPyRelation:
        """DuckDB relation object"""
        return self._rel

    def native_as_df(self) -> DuckDBPyRelation:
        return self._rel

    @property
    def empty(self) -> bool:
        return self.count() == 0

    def peek_array(self) -> List[Any]:
        res = self._rel.limit(1).to_df()
        if res.shape[0] == 0:
            raise FugueDatasetEmptyError()
        return res.values.tolist()[0]

    def count(self) -> int:
        return len(self._rel)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        return DuckDataFrame(_drop_duckdb_columns(self._rel, cols))

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        return DuckDataFrame(_select_duckdb_columns(self._rel, keys))

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        _assert_no_missing(self._rel, columns.keys())
        expr = ", ".join(
            f"{a} AS {b}"
            for a, b in [
                (encode_column_name(name), encode_column_name(columns.get(name, name)))
                for name in self._rel.columns
            ]
        )
        return DuckDataFrame(self._rel.project(expr))

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self.schema.alter(columns)
        if new_schema == self.schema:
            return self
        fields: List[str] = []
        for f1, f2 in zip(self.schema.fields, new_schema.fields):
            if f1.type == f2.type:
                fields.append(encode_column_name(f1.name))
            else:
                tp = to_duck_type(f2.type)
                fields.append(
                    f"CAST({encode_column_name(f1.name)} AS {tp}) "
                    f"AS {encode_column_name(f1.name)}"
                )
        return DuckDataFrame(self._rel.project(", ".join(fields)))

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return _duck_as_arrow(self._rel)

    def as_pandas(self) -> pd.DataFrame:
        return _duck_as_pandas(self._rel)

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        res = ArrowDataFrame(self.as_arrow())
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return _duck_as_array(self._rel, columns=columns, type_safe=type_safe)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        yield from _duck_as_array_iterable(
            self._rel, columns=columns, type_safe=type_safe
        )

    def as_dicts(self, columns: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return _duck_as_dicts(self._rel, columns=columns)

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        yield from _duck_as_dict_iterable(self._rel, columns=columns)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        if columns is not None:
            return self[columns].head(n)
        return ArrowDataFrame(_duck_as_arrow(self._rel.limit(n)))


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, DuckDBPyRelation))
def _duckdb_as_fugue_df(df: DuckDBPyRelation, **kwargs: Any) -> DuckDataFrame:
    return DuckDataFrame(df, **kwargs)


@is_df.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_is_df(df: DuckDBPyRelation) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duckdb_num_partitions(df: DuckDBPyRelation) -> int:
    return 1


@as_local_bounded.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_as_local(df: DuckDBPyRelation) -> DuckDBPyRelation:
    return df


@as_arrow.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_as_arrow(df: DuckDBPyRelation) -> pa.Table:
    _df = df.fetch_arrow_table()
    _df = replace_types_in_table(_df, LARGE_TYPES_REPLACEMENT, recursive=True)
    return _df


@as_pandas.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_as_pandas(df: DuckDBPyRelation) -> pd.DataFrame:
    adf = _duck_as_arrow(df)
    return _pa_table_as_pandas(adf)


@get_schema.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_get_schema(df: DuckDBPyRelation) -> Schema:
    return Schema([pa.field(x, to_pa_type(y)) for x, y in zip(df.columns, df.types)])


@get_column_names.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _get_duckdb_columns(df: DuckDBPyRelation) -> List[Any]:
    return list(df.columns)


@select_columns.candidate(lambda df, *args, **kwargs: isinstance(df, DuckDBPyRelation))
def _select_duckdb_columns(
    df: DuckDBPyRelation, columns: List[Any]
) -> DuckDBPyRelation:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns)
    return df.project(",".join(encode_column_name(n) for n in columns))


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, DuckDBPyRelation))
def _drop_duckdb_columns(df: DuckDBPyRelation, columns: List[str]) -> DuckDBPyRelation:
    # if len(columns) == 0:
    #   return df
    _columns = {c: 1 for c in columns}
    cols = [col for col in df.columns if _columns.pop(col, None) is None]
    assert_or_throw(
        len(cols) > 0, FugueDataFrameOperationError("must keep at least one column")
    )
    assert_or_throw(
        len(_columns) == 0,
        FugueDataFrameOperationError("found nonexistent columns {_columns}"),
    )
    return df.project(",".join(encode_column_name(n) for n in cols))


@as_array.candidate(lambda df, *args, **kwargs: isinstance(df, DuckDBPyRelation))
def _duck_as_array(
    df: DuckDBPyRelation, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Any]:
    return pa_table_as_array(df.fetch_arrow_table(), columns=columns)


@as_array_iterable.candidate(
    lambda df, *args, **kwargs: isinstance(df, DuckDBPyRelation)
)
def _duck_as_array_iterable(
    df: DuckDBPyRelation, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Any]:
    yield from pa_table_as_array_iterable(df.fetch_arrow_table(), columns=columns)


@as_dicts.candidate(lambda df, *args, **kwargs: isinstance(df, DuckDBPyRelation))
def _duck_as_dicts(
    df: DuckDBPyRelation, columns: Optional[List[str]] = None
) -> List[Dict[str, Any]]:
    return pa_table_as_dicts(df.fetch_arrow_table(), columns=columns)


@as_dict_iterable.candidate(
    lambda df, *args, **kwargs: isinstance(df, DuckDBPyRelation)
)
def _duck_as_dict_iterable(
    df: DuckDBPyRelation, columns: Optional[List[str]] = None
) -> Iterable[Dict[str, Any]]:
    yield from pa_table_as_dict_iterable(df.fetch_arrow_table(), columns=columns)


def _assert_no_missing(df: DuckDBPyRelation, columns: Iterable[Any]) -> None:
    missing = set(columns) - set(df.columns)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")



from typing import Any, Dict

import ray
from packaging import version

FUGUE_RAY_CONF_SHUFFLE_PARTITIONS = "fugue.ray.shuffle.partitions"
FUGUE_RAY_DEFAULT_PARTITIONS = "fugue.ray.default.partitions"
FUGUE_RAY_DEFAULT_BATCH_SIZE = "fugue.ray.default.batch_size"
FUGUE_RAY_ZERO_COPY = "fugue.ray.zero_copy"

FUGUE_RAY_DEFAULT_CONF: Dict[str, Any] = {
    FUGUE_RAY_CONF_SHUFFLE_PARTITIONS: -1,
    FUGUE_RAY_DEFAULT_PARTITIONS: 0,
    FUGUE_RAY_ZERO_COPY: True,
}
RAY_VERSION = version.parse(ray.__version__)

_ZERO_COPY: Dict[str, Any] = {"zero_copy_batch": True}



from typing import Any, Callable, Optional

from triad import Schema, assert_or_throw, to_uuid

from fugue.collections.yielded import Yielded
from fugue.dataframe import DataFrame
from fugue.exceptions import FugueWorkflowCompileError
from fugue.execution.api import as_fugue_engine_df
from fugue.extensions.creator import Creator


class Load(Creator):
    def create(self) -> DataFrame:
        kwargs = self.params.get("params", dict())
        path = self.params.get_or_throw("path", str)
        format_hint = self.params.get("fmt", "")
        columns = self.params.get_or_none("columns", object)

        return self.execution_engine.load_df(
            path=path, format_hint=format_hint, columns=columns, **kwargs
        )


class CreateData(Creator):
    def __init__(
        self,
        df: Any,
        schema: Any = None,
        data_determiner: Optional[Callable[[Any], Any]] = None,
    ) -> None:
        if isinstance(df, Yielded):
            assert_or_throw(
                schema is None,
                FugueWorkflowCompileError("schema must be None when data is Yielded"),
            )
        super().__init__()
        self._df = df
        self._schema = schema if schema is None else Schema(schema)
        self._data_determiner = data_determiner

    def create(self) -> DataFrame:
        if isinstance(self._df, Yielded):
            return self.execution_engine.load_yielded(self._df)
        return as_fugue_engine_df(self.execution_engine, self._df, schema=self._schema)

    def _df_uid(self):
        if self._data_determiner is not None:
            return self._data_determiner(self._df)
        if isinstance(self._df, Yielded):
            return self._df
        return 1

    def __uuid__(self) -> str:
        return to_uuid(super().__uuid__(), self._df_uid(), self._schema)



# flake8: noqa
from fugue import register_execution_engine, register_sql_engine

from fugue_duckdb.execution_engine import DuckDBEngine, DuckExecutionEngine

try:
    from fugue_duckdb.dask import DuckDaskExecutionEngine
except Exception:  # pragma: no cover
    pass
